
"""
This module provides :py:mod:`Django` views
exposing a :py:class:`livemetrics.LiveMetrics` object.

A minimal application would be:

.. code-block:: python

    import livemetrics
    import livemetrics.publishers.django
    
    import django
    from django.http import HttpResponse
    from django.urls import path

    # Declare a global LiveMetrics object in one of your modules
    LM = livemetrics.LiveMetrics(version,about,is_healthy,is_ready)

    @LM.timer('sample','ok','error')
    def sample_view(request):
        return HttpResponse("sample")

In your :file:`urls.py`, register your views with:

.. code-block:: python

    urlpatterns = [
        path('sample', sample_view, name='sample'),
    ]

    urlpatterns += livemetrics.publishers.django.urlpatterns(LM)

"""

import json

from django.http import HttpResponse
from django.views import View
from django.urls import path

class About(View):
    LM = None
    def get(self,request):
        return HttpResponse(self.LM.about)

class IsHealthy(View):
    LM = None
    def get(self,request):
        status = self.LM.is_healthy()
        if status:
            resp = HttpResponse('OK')
            resp["Access-Control-Allow-Origin"] = "*"
            return resp
        else:
            resp = HttpResponse('KO', status=500)
            resp["Access-Control-Allow-Origin"] = "*"
            return resp

class IsReady(View):
    LM = None
    def get(self,request):
        status = self.LM.is_ready()
        if status:
            resp = HttpResponse('OK')
            resp["Access-Control-Allow-Origin"] = "*"
            return resp
        else:
            resp = HttpResponse('KO', status=500)
            resp["Access-Control-Allow-Origin"] = "*"
            return resp

class Version(View):
    LM = None
    def get(self,request):
        resp = HttpResponse(self.LM.version,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Meters3(View):
    LM =  None
    def get(self,request,event,result,metric):
        data = self.LM.get_metrics(event,result,metric)
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Meters2(View):
    LM =  None
    def get(self,request,event,result):
        data = self.LM.get_metrics(event,result)
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Meters1(View):
    LM =  None
    def get(self,request,event):
        data = self.LM.get_metrics(event)
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Meters0(View):
    LM =  None
    def get(self,request):
        data = self.LM.get_metrics()
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Gauges2(View):
    LM =  None
    def get(self,request,object,metric):
        data = self.LM.get_gauges(object,metric)
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Gauges1(View):
    LM =  None
    def get(self,request,object):
        data = self.LM.get_gauges(object)
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

class Gauges0(View):
    LM =  None
    def get(self,request):
        data = self.LM.get_gauges()
        data = json.dumps(data)
        resp = HttpResponse(data,content_type='application/json')
        resp["Access-Control-Allow-Origin"] = "*"
        return resp

def _get_histograms(LM,request,event,metric):
    msg = None
    params = {}
    params.update(request.GET)
    if not isinstance(params.setdefault('percentiles',[0.05, 0.25, 0.5, 0.75, 0.95]),list):
        params['percentiles'] = [params['percentiles']]
    if not isinstance(params.setdefault('scale',[10]),list):
        params['scale'] = [params['scale']]
    if len(params)!=2:
        keys = ", ".join(sorted(list(set(params.keys()) - {'percentiles', 'scale'})))
        msg="Invalid query parameters: " + keys
    if not metric in [None,'count','min','max','mean','stddev','quantiles','distribution']:
        msg = "Unknown metric " + metric
    if msg:
        return HttpResponse(msg, status=400)

    # Make sure the types are correct
    params['percentiles'] = [float(x) for x in params['percentiles']]
    params['scale'] = int(params['scale'][0])

    try:
        data = LM.get_histograms(event,metric,**params)
    except Exception as exc:
        msg = str(exc)
        return HttpResponse(msg, status=500)
    data = json.dumps(data)
    resp = HttpResponse(data,content_type='application/json')
    resp["Access-Control-Allow-Origin"] = "*"
    return resp

class Histograms2(View):
    LM =  None
    def get(self,request,event,metric):
        return _get_histograms(self.LM,request,event,metric)

class Histograms1(View):
    LM =  None
    def get(self,request,event):
        return _get_histograms(self.LM,request,event,None)

class Histograms0(View):
    LM =  None
    def get(self,request):
        return _get_histograms(self.LM,request,None,None)

def urlpatterns(LM):
    """
    Return a list of Django paths to be registered in the application.

    *LM*: a :py:class:`livemetrics.LiveMetrics` object
    """
    urlpatterns = [
        path('monitoring/v1/about', About.as_view(LM=LM),name='monitoring-about'),
        path('monitoring/v1/is_healthy', IsHealthy.as_view(LM=LM),name='monitoring-is_healthy'),
        path('monitoring/v1/is_ready', IsReady.as_view(LM=LM),name='monitoring-is_ready'),
        path('monitoring/v1/version', Version.as_view(LM=LM),name='monitoring-version'),

        path('monitoring/v1/metrics/meters/<event>/<result>/<metric>', Meters3.as_view(LM=LM),name='monitoring-meters3'),
        path('monitoring/v1/metrics/meters/<event>/<result>', Meters2.as_view(LM=LM),name='monitoring-meters2'),
        path('monitoring/v1/metrics/meters/<event>', Meters1.as_view(LM=LM),name='monitoring-meters1'),
        path('monitoring/v1/metrics/meters', Meters0.as_view(LM=LM),name='monitoring-meters0'),

        path('monitoring/v1/metrics/gauges/<object>/<metric>', Gauges2.as_view(LM=LM),name='monitoring-gauges2'),
        path('monitoring/v1/metrics/gauges/<object>', Gauges1.as_view(LM=LM),name='monitoring-gauges1'),
        path('monitoring/v1/metrics/gauges', Gauges0.as_view(LM=LM),name='monitoring-gauges0'),

        path('monitoring/v1/metrics/histograms/<event>/<metric>', Histograms2.as_view(LM=LM),name='monitoring-histograms2'),
        path('monitoring/v1/metrics/histograms/<event>', Histograms1.as_view(LM=LM),name='monitoring-histograms1'),
        path('monitoring/v1/metrics/histograms', Histograms0.as_view(LM=LM),name='monitoring-histograms0'),
    ]
    return urlpatterns




import http.server
import random
import time

import livemetrics
import livemetrics.publishers.http

LM = livemetrics.LiveMetrics('{"version":"1.0"}',"Test server",True, True)

LM.gauge("random",lambda: random.randint(1,10))

#______________________________________________________________________________
class TestHTTPRequestHandler(livemetrics.publishers.http.HTTPRequestHandler):

    @LM.timer("time",None,None)
    def do_test(self):
        value = int(self.path.split('/')[-1])
        LM.mark('test','ok')
        if value % 5 == 0:
            LM.mark('test','error')
        LM.histogram("values",value)
        time.sleep(random.triangular(0.1,0.2,0.4))
        self.send_response(200)
        self.end_headers()

    def do_GET(self):
        if self.path.startswith('/test'):
            return self.do_test()
        else:
            return super().do_GET()

    def log_message(self, format, *args):
        pass

def _serve():
    global LM
    server_address = ('0.0.0.0', 7070)
    httpd = http.server.HTTPServer(server_address, lambda r,a,s: TestHTTPRequestHandler(LM,r,a,s))
    httpd.allow_reuse_address = True
    httpd.serve_forever()

# ______________________________________________________________________________
if __name__=='__main__':
    _serve()





import livemetrics

project = 'livemetrics'
copyright = livemetrics.__copyright__
version = livemetrics.__version__
release = livemetrics.__version__

master_doc = 'index'

html_static_path = ['_static']
html_theme = 'alabaster'
html_theme_options = {
    'github_user': 'idemia',
    'github_repo': 'python-livemetrics',
    'github_button': True,
    'show_relbars': True,
    'fixed_sidebar': False,
}
extensions = ['sphinx.ext.autodoc']




import os
import time
import unittest
import requests

IP = '127.0.0.1'
PORT = '8765'

#_______________________________________________________________________________
class TestPublisher(unittest.TestCase):
    
    def test_a(self):
        # Send some events
        requests.get('http://'+IP+':'+PORT+'/test')
        requests.get('http://'+IP+':'+PORT+'/test')
        requests.get('http://'+IP+':'+PORT+'/test')
        requests.get('http://'+IP+':'+PORT+'/test')
        requests.get('http://'+IP+':'+PORT+'/test')

        # Get the application status
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/about') as r:
            self.assertEqual(b'Test server',r.content)
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/version') as r:
            self.assertEqual('1.0',r.json()['version'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/is_healthy') as r:
            self.assertEqual(200,r.status_code)
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/is_ready') as r:
            self.assertEqual(200,r.status_code)

        # Meters
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters') as r:
            self.assertEqual(5,r.json()['test']['ok']['count'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters/test') as r:
            self.assertEqual(5,r.json()['ok']['count'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters/test/ok') as r:
            self.assertEqual(5,r.json()['count'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters/test/ok/count') as r:
            self.assertEqual(5,r.json())
        # Missing meter
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters/missing/ok/count') as r:
            self.assertEqual(0,r.json())
    
        # Gauges
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges') as r:
            self.assertEqual(10,r.json()['fixed']['count'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/fixed') as r:
            self.assertEqual(10,r.json()['count'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/fixed/count') as r:
            self.assertEqual(10,r.json())
        # Missing gauge
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/missing/count') as r:
            self.assertEqual(0,r.json())
    
        # Histograms
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms') as r:
            self.assertGreater(20,r.json()['histo']['quantiles']['0.5'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo') as r:
            self.assertGreater(20,r.json()['quantiles']['0.5'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/quantiles?percentiles=0.5') as r:
            self.assertGreater(20,r.json()['0.5'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/distribution?scale=1') as r:
            self.assertGreater(r.json()[0],4)
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/count') as r:
            self.assertEqual(5,r.json())
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/count?bad_param=2') as r:
            self.assertEqual(400,r.status_code)
            self.assertEqual(b"Invalid query parameters: bad_param",r.content)
        # Missing histogram
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/missing/count') as r:
            self.assertEqual(0,r.json())
    
        # Histograms with errors
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/quantiles?bad_param=0.5&bad2=1') as r:
            self.assertEqual(400,r.status_code)
            self.assertEqual(b"Invalid query parameters: bad2, bad_param",r.content)
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/quantiles?percentiles=0.5&percentiles=0.75&bad_param=0.5') as r:
            self.assertEqual(400,r.status_code)
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/histograms/histo/bad_metric') as r:
            self.assertEqual(400,r.status_code)

        # Test the decoration works
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters/decorator_async/ok') as r:
            self.assertEqual(4,r.json()['count'])
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/meters/decorator_async/error') as r:
            self.assertEqual(1,r.json()['count'])

        # Additional error cases
        with requests.get('http://'+IP+':'+PORT+'/bad/path/v1/metrics/histograms/histo/bad_metric') as r:
            self.assertEqual(404,r.status_code)

        # Test with a bad status
        backup_ih = self.LM.is_healthy
        self.LM.is_healthy = lambda: False

        backup_ir = self.LM.is_ready
        self.LM.is_ready = lambda: False

        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/is_healthy') as r:
            self.assertEqual(500,r.status_code)
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/is_ready') as r:
            self.assertEqual(500,r.status_code)

        self.LM.is_healthy = backup_ih
        self.LM.is_ready = backup_ir

        # Test memory & CPU
        if not os.name=='posix':
            return

        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/memory/count') as r:
            self.assertGreater(r.json(),20000000)

        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/cpu/count') as r:
            cpu = r.json()
        with requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/cpu/count') as r:
            self.assertEqual(cpu,r.json())
        time.sleep(0.5)
        requests.get('http://'+IP+':'+PORT+'/monitoring/v1/metrics/gauges/cpu/count')




"""
This module provides a :py:mod:`flask` blueprint
exposing a :py:class:`livemetrics.LiveMetrics` object.

A minimal application would be:

.. code-block:: python

    import livemetrics
    import livemetrics.publishers.flask
    
    from flask import Flask
    app = Flask(__name__)

    LM = livemetrics.LiveMetrics(version,about,is_healthy,is_ready)

    @app.route("/sample")
    @LM.timer('sample','ok','error')
    def sample():
        return "Sample!"

    app.register_blueprint(livemetrics.publishers.flask.blueprint(LM))

"""

import json

from flask import Blueprint, make_response, request

class Handler:
    def __init__(self,LM):
        self.LM = LM

    def about(self):
        data = self.LM.about
        resp = make_response(data, 200)
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def is_healthy(self):
        status = self.LM.is_healthy()
        if status:
            resp = make_response("OK", 200)
            resp.headers['Access-Control-Allow-Origin'] = '*'
            return resp
        else:
            resp = make_response("KO", 500)
            resp.headers['Access-Control-Allow-Origin'] = '*'
            return resp

    def is_ready(self):
        status = self.LM.is_ready()
        if status:
            resp = make_response("OK", 200)
            resp.headers['Access-Control-Allow-Origin'] = '*'
            return resp
        else:
            resp = make_response("KO", 500)
            resp.headers['Access-Control-Allow-Origin'] = '*'
            return resp

    def version(self):
        data = self.LM.version
        resp = make_response(data, 200)
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_meters3(self,event,result,metric):
        data = self.LM.get_metrics(event,result,metric)
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_meters2(self,event,result):
        data = self.LM.get_metrics(event,result)
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_meters1(self,event):
        data = self.LM.get_metrics(event)
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_meters0(self):
        data = self.LM.get_metrics()
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_gauges2(self,object,metric):
        data = self.LM.get_gauges(object,metric)
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_gauges1(self,object):
        data = self.LM.get_gauges(object)
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_gauges0(self):
        data = self.LM.get_gauges()
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def _get_histograms(self,event,metric):
        msg = None
        params = {}
        params.update(request.args)
        if not isinstance(params.setdefault('percentiles',[0.05, 0.25, 0.5, 0.75, 0.95]),list):
            params['percentiles'] = [params['percentiles']]
        if not isinstance(params.setdefault('scale',[10]),list):
            params['scale'] = [params['scale']]

        if len(params)!=2:
            keys = ", ".join(sorted(list(set(params.keys()) - {'percentiles', 'scale'})))
            msg="Invalid query parameters: " + keys
        if not metric in [None,'count','min','max','mean','stddev','quantiles','distribution']:
            msg = "Unknown metric " + metric
        if msg:
            return make_response(msg, 400)

        # Make sure the types are correct
        params['percentiles'] = [float(x) for x in params['percentiles']]
        params['scale'] = int(params['scale'][0])

        try:
            data = self.LM.get_histograms(event,metric,**params)
        except Exception as exc:
            msg = str(exc)
            return make_response(msg, 500)
        data = json.dumps(data)
        resp = make_response(data, 200)
        resp.headers['Content-Length'] = str(len(data))
        resp.headers['Content-Type'] = 'application/json'
        resp.headers['Access-Control-Allow-Origin'] = '*'
        return resp

    def get_histograms2(self,event,metric):
        return self._get_histograms(event,metric)

    def get_histograms1(self,event):
        return self._get_histograms(event,None)

    def get_histograms0(self):
        return self._get_histograms(None,None)

def blueprint(LM):
    """
    Return a blueprint with the routes and view_func for the publication of the metrics.

    *LM*: a :py:class:`livemetrics.LiveMetrics` object
    """
    handler = Handler(LM)
    blueprint = Blueprint('livemetrics', __name__)
    blueprint.add_url_rule('/monitoring/v1/about',view_func=handler.about)
    blueprint.add_url_rule('/monitoring/v1/is_healthy',view_func=handler.is_healthy)
    blueprint.add_url_rule('/monitoring/v1/is_ready',view_func=handler.is_ready)
    blueprint.add_url_rule('/monitoring/v1/version',view_func=handler.version)

    blueprint.add_url_rule('/monitoring/v1/metrics/meters/<event>/<result>/<metric>',view_func=handler.get_meters3)
    blueprint.add_url_rule('/monitoring/v1/metrics/meters/<event>/<result>',view_func=handler.get_meters2)
    blueprint.add_url_rule('/monitoring/v1/metrics/meters/<event>',view_func=handler.get_meters1)
    blueprint.add_url_rule('/monitoring/v1/metrics/meters',view_func=handler.get_meters0)

    blueprint.add_url_rule('/monitoring/v1/metrics/gauges/<object>/<metric>',view_func=handler.get_gauges2)
    blueprint.add_url_rule('/monitoring/v1/metrics/gauges/<object>',view_func=handler.get_gauges1)
    blueprint.add_url_rule('/monitoring/v1/metrics/gauges',view_func=handler.get_gauges0)

    blueprint.add_url_rule('/monitoring/v1/metrics/histograms/<event>/<metric>',view_func=handler.get_histograms2)
    blueprint.add_url_rule('/monitoring/v1/metrics/histograms/<event>',view_func=handler.get_histograms1)
    blueprint.add_url_rule('/monitoring/v1/metrics/histograms',view_func=handler.get_histograms0)

    return blueprint





"""

This module provides a (very) minimal application to demonstrate the usage of metrics.

Install with::

    pip install livemetrics[dashboard]

Run with::

    livemetrics-dashboard -i myconf.yaml

And then access http://localhost:9000/ to visualize the dashboard.

The configuration file describes rows, and for each row the cells to display. Each cell is associated
to one value (or to the statistics of one value for histograms).

The configuration file is a Jinja2 template used to generate a YAML file.
All macros and filters of Jinja2 are available, making possible to define template or macros for
complex cases.

A simple example:

.. code-block:: jinja

    {% set server = "http://localhost:7070/monitoring/v1" %}
    ---
    title: Sample
    rows:
    - label: First Line
      cells:
      - label: Line 1
        type: label
      - label: Health
        server: "{{server}}/is_healthy"
        type: status
      - label: CPU
        server: "{{server}}/metrics/gauges/cpu/count"
        type: value
        unit: "%"
        factor: "*100"
        color: green
      - label: Memory
        server: "{{server}}/metrics/gauges/memory/count"
        type: gauge
        unit: MB
        factor: "/1048576"
        max: 256
        gauge_color: blue
        color: "#C0D0F0"
      - type: empty
      - label: Time
        server: "{{server}}/metrics/histograms/time"
        type: histogram
        unit: ms
        factor: "*1000"
        color: "#ff000080"
        median_color: green
      - label: Test
        server: "{{server}}/metrics/meters/test/ok/rate1"
        type: gauge
        unit: "/h"
        factor: "*3600"
        gauge_color: green
        color: "#C0F0C0"
        max: 25000

The cells can be of the following types:

.. list-table::
    :widths: 10 40 50

    * - Type
      - Description
      - Options

    * - ``empty``
      - Display an empty cell
      - - N/A
    * - ``label``
      - Display a static text
      - - ``label``: The text to display
    * - ``status``
      - Display a boolean value as a colored icon
      - - ``label``: A text displayed below the icon
        - ``server``: The full URL used to get the boolean value
    * - ``value``
      - Display a numeric value as a text
      - - ``label``: A text displayed below the value
        - ``server``: The full URL used to get the value
        - ``unit``: A small text added after the value
        - ``factor``: A factor to apply to the value. It must start with ``*`` or ``/`` to indicate the operation to execute
        - ``color``: The text color
    * - ``gauge``
      - Display a numeric value as a gauge and a sparkline showing the recent evolution of the value
      - - ``label``: A text displayed below the value
        - ``server``: The full URL used to get the value
        - ``unit``: A small text added after the label
        - ``factor``: A factor to apply to the value. It must start with ``*`` or ``/`` to indicate the operation to execute
        - ``color``: The fill color of the sparkline
        - ``gauge_color``: The color of the gauge
        - ``max``: The maximum value of the gauge
    * - ``histogram``
      - Display statistics about a value as a whisker box
      - - ``label``: A text displayed below the value
        - ``server``: The full URL used to get the statistics. It must return a json dictionary
          with values for the keys ["0.05", "0.25", "0.75", "0.95", "0.5"]
        - ``unit``: A small text added as the legend of the diagram
        - ``factor``: A factor to apply to the value. It must start with ``*`` or ``/`` to indicate the operation to execute
        - ``color``: The fill color of the box
        - ``median_color``: The color of the line showing the median value

"""

import os
import logging
import argparse

import socketserver
import http.server
import urllib.request
import pickle
import json
import yaml
from urllib.parse import urlparse,parse_qs

from jinja2 import Template,Environment,FileSystemLoader,PrefixLoader,ChoiceLoader

class DashBoardHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    conf = None

    def __init__(self, request, client_address, server, options):
        if DashBoardHTTPRequestHandler.conf is None:
            DashBoardHTTPRequestHandler.conf = 1

            myself = {'self':     FileSystemLoader(os.path.dirname(__file__)) }
            loader = PrefixLoader(myself)

            env = Environment(loader=ChoiceLoader([FileSystemLoader([".",'/']), loader]))
            template = env.get_template(options.input.name)
            buf = template.render()

            DashBoardHTTPRequestHandler.conf = yaml.load(buf,Loader=yaml.Loader)

            # generate ids
            i = 1
            for r in DashBoardHTTPRequestHandler.conf['rows']:
                for c in r['cells']:
                    c['id'] = str("id%d"%i)
                    i += 1

        super().__init__(request, client_address, server)

    def end_headers(self):
        self.send_header("Access-Control-Allow-Origin","*")
        return http.server.SimpleHTTPRequestHandler.end_headers(self)

    def do_GET(self):
        if self.path=='/':
            myself = {'self':     FileSystemLoader(os.path.dirname(__file__)) }
            loader = PrefixLoader(myself)
            env = Environment(loader=loader)

            template = env.get_template('self/dashboard.html')
            buf = template.render(config=self.conf)
            self.send_response(200)
            self.send_header("Content-Type","text/html ; encoding=UTF-8")
            self.send_header("Content-Length", str(len(buf)))
            self.end_headers()
            self.wfile.write(buf.encode('utf-8'))
            return
        elif self.path.startswith('/all'):
            pr = urlparse(self.path)
            params = parse_qs(pr.query)
            try:
                with urllib.request.urlopen(params['server'][0]) as obj:
                    buf = obj.read()
                    self.send_response(obj.getcode())
                    if 'content-type' in obj.info():
                        self.send_header("Content-Type",obj.info()['content-type'])
                    self.send_header("Content-Length", str(len(buf)))
                    self.end_headers()
                    self.wfile.write(buf)
                    return
            except:
                self.send_response(500)
                self.end_headers()
                return
        return super().do_GET()

    def log_message(self, format, *args):
        # override to include in the standard log file and to apply filtering
        logging.debug("%s - - [%s] %s" %
                         (self.client_address[0],
                          self.log_date_time_string(),
                          format%args))

class ThreadedHTTPServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    """Handle requests in a separate thread."""

def serve(options):
    server_address = (options.ip, options.port)
    httpd = http.server.HTTPServer(server_address, lambda r,a,s:  DashBoardHTTPRequestHandler(r,a,s,options))
    httpd.serve_forever()

#______________________________________________________________________________
def main(args=None):

    parser = argparse.ArgumentParser(description='Dashboard Sample')

    parser.add_argument("-i", "--input",required=True,dest='input',type=argparse.FileType(mode='r'),help="Configuration file")
    parser.add_argument("-I", "--listening-ip",default='0.0.0.0',dest='ip',help="Listening IP address (default: 0.0.0.0)")
    parser.add_argument("-P", "--listening-port",default=9000,dest='port',type=int,help="Listening port number (default: 9000)")
    parser.add_argument("-l", "--loglevel",default='INFO',dest='loglevel',help="Log level")
    parser.add_argument("-f", "--logfile",default=None,dest='logfile',help="Log file")

    options = parser.parse_args(args=args)

    logging.basicConfig(format='%(message)s',level=logging.getLevelName(options.loglevel))
    if options.logfile:
        fh = logging.handlers.TimedRotatingFileHandler(options.logfile,when='H',interval=8)
        fh.setLevel(logging.getLevelName(options.loglevel))
        formatter = logging.Formatter('%(asctime)-15s %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        logging.getLogger().addHandler(fh)

    logging.debug(options)

    serve(options)

if __name__=='__main__':
    main()




"""
This module provides a :py:class:`http.server.BaseHTTPRequestHandler` subclass
exposing a :py:class:`livemetrics.LiveMetrics` object.

A minimal application would be:

.. code-block:: python

    import livemetrics
    import livemetrics.publishers.http

    import http.server

    LM = livemetrics.LiveMetrics(version,about,is_healthy,is_ready)

    class Sample(livemetrics.publishers.http.HTTPRequestHandler):

        @LM.timer("sample","ok","error")
        def do_GET(self):
            if self.path=='/sample':
                self.send_response(200)
                self.end_headers()
            else:
                return super().do_GET()

    httpd = http.server.HTTPServer((host,port), lambda r,a,s: Sample(LM,r,a,s))
    httpd.serve_forever()

"""

import json
import http.server
import posixpath
from urllib.parse import unquote
from urllib.parse import urlparse,parse_qs

#______________________________________________________________________________
class HTTPRequestHandler(http.server.BaseHTTPRequestHandler):
    """
    Subclass this class in your application to inherit the handling of GET requests
    exposing a :py:class:`livemetrics.LiveMetrics` object.
    """

    def __init__(self,LM,request,client_address,server):
        self.LM = LM
        return super().__init__(request,client_address,server)

    def do_GET(self):
        if self.path=='/monitoring/v1/about':
            data = self.LM.about
            self.send_response(200)
            self.send_header("Content-Type","text/plain")
            self.send_header("Access-Control-Allow-Origin","*")
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data.encode('ascii'))
        elif self.path=='/monitoring/v1/is_healthy':
            status = self.LM.is_healthy()
            if status:
                data = 'OK'
                self.send_response(200)
                self.send_header("Content-Type","text/plain")
                self.send_header("Access-Control-Allow-Origin","*")
                self.send_header("Content-Length", str(len(data)))
                self.end_headers()
                self.wfile.write(data.encode('ascii'))
            else:
                data = 'KO'
                self.send_response(500)
                self.send_header("Content-Type","text/plain")
                self.send_header("Access-Control-Allow-Origin","*")
                self.send_header("Content-Length", str(len(data)))
                self.end_headers()
                self.wfile.write(data.encode('ascii'))
        elif self.path=='/monitoring/v1/is_ready':
            status = self.LM.is_ready()
            if status:
                data = 'OK'
                self.send_response(200)
                self.send_header("Content-Type","text/plain")
                self.send_header("Access-Control-Allow-Origin","*")
                self.send_header("Content-Length", str(len(data)))
                self.end_headers()
                self.wfile.write(data.encode('ascii'))
            else:
                data = 'KO'
                self.send_response(500)
                self.send_header("Content-Type","text/plain")
                self.send_header("Access-Control-Allow-Origin","*")
                self.send_header("Content-Length", str(len(data)))
                self.end_headers()
                self.wfile.write(data.encode('ascii'))
        elif self.path=='/monitoring/v1/version':
            data = self.LM.version
            self.send_response(200)
            self.send_header("Content-Type","application/json")
            self.send_header("Access-Control-Allow-Origin","*")
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data.encode('ascii'))
        elif self.path.startswith('/monitoring/v1/metrics/meters'):
            # Parse path {event}/{result}/{metric}
            path = posixpath.normpath(unquote(self.path)).split('/')
            path.extend([None,None,None])
            event,result,metric = path[5:8]
            self.send_response(200)
            self.send_header("Content-Type","application/json")
            self.send_header("Access-Control-Allow-Origin","*")
            data = self.LM.get_metrics(event,result,metric)
            data = json.dumps(data)
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data.encode('ascii'))
        elif self.path.startswith('/monitoring/v1/metrics/gauges'):
            # Parse path {object}/{metric}
            path = posixpath.normpath(unquote(self.path)).split('/')
            path.extend([None,None])
            name,metric = path[5:7]
            self.send_response(200)
            self.send_header("Content-Type","application/json")
            self.send_header("Access-Control-Allow-Origin","*")
            data = self.LM.get_gauges(name,metric)
            data = json.dumps(data)
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data.encode('ascii'))
        elif self.path.startswith('/monitoring/v1/metrics/histograms'):
            pr = urlparse(self.path)
            # Parse path {event}/{metric}
            path = posixpath.normpath(unquote(pr.path)).split('/')
            path.extend([None,None])
            event,metric = path[5:7]
            # analyze query parameters
            params = parse_qs(pr.query)
            if 'percentiles' not in params:
                # Define default value
                params['percentiles'] = [0.05, 0.25, 0.5, 0.75, 0.95]
            if 'scale' not in params:
                # Define default value
                params['scale'] = [10]
            msg = None
            if len(params)!=2:
                keys = ", ".join(sorted(list(set(params.keys()) - {'percentiles', 'scale'})))
                msg="Invalid query parameters: " + keys
            if not metric in [None,'count','min','max','mean','stddev','quantiles','distribution']:
                msg = "Unknown metric " + metric
            if msg:
                self.send_response(400)
                self.send_header("Access-Control-Allow-Origin","*")
                self.send_header("Content-Length", str(len(msg)))
                self.end_headers()
                self.wfile.write(msg.encode('ascii'))
                return
            # Make sure the types are correct
            params['percentiles'] = [float(x) for x in params['percentiles']]
            params['scale'] = int(params['scale'][0])
            try:
                data = self.LM.get_histograms(event,metric,**params)
            except Exception as exc:
                msg = str(exc)
                self.send_response(500)
                self.send_header("Access-Control-Allow-Origin","*")
                self.send_header("Content-Length", str(len(msg)))
                self.end_headers()
                self.wfile.write(msg.encode('ascii'))
                return
            data = json.dumps(data)
            self.send_response(200)
            self.send_header("Content-Type","application/json")
            self.send_header("Access-Control-Allow-Origin","*")
            self.send_header("Content-Length", str(len(data)))
            self.end_headers()
            self.wfile.write(data.encode('ascii'))
        else:
            self.send_response(404)
            self.send_header("Access-Control-Allow-Origin","*")
            self.end_headers()




import threading
import unittest
import random
import time

import django
from django.http import HttpResponse
from django.urls import path
from django.core.management import call_command

import livemetrics
import livemetrics.publishers.django

import tests.publishers

LM = livemetrics.LiveMetrics('{"version":"1.0"}',"Test server",True, True)
# Sample of gauges
LM.gauge('fixed',10)
LM.gauge('test',10)
LM.gauge('call',lambda: random.randrange(100,1000))

#______________________________________________________________________________
def ok(ret):
    return str(ret.status)

def error(exc):
    return str(exc)

def exc(ret):
    raise Exception("Bad status")

COUNT = 0

@LM.timer("decorator_async","ok","error")
@LM.timer("decorator_async_callable",ok,error)
@LM.timer("decorator_async_histo_only",None,None)
@LM.timer("decorator_async_callable_exc",exc,exc)
def myview(request):
    time.sleep(random.random()/10.)
    global LM
    LM.mark('test',"ok")
    LM.gauge('test',random.randrange(10,10000))
    LM.histogram("histo",random.random()*20)
    global COUNT
    COUNT += 1
    if COUNT % 3 ==0:
        raise Exception("false-alert")
    return HttpResponse("OK")

urlpatterns = [
    path('test', myview, name='test'),
]

urlpatterns += livemetrics.publishers.django.urlpatterns(LM)

def _serve():
    from django.conf import settings

    global LM
    tests.publishers.PORT = '8768'

    if not settings.configured:
        # Configure test environment
        settings.configure(
            DATABASES={
                'default': {
                    'ENGINE': 'django.db.backends.sqlite3',
                    'NAME': ':memory:'
                }
            },
            INSTALLED_APPS=(
                'tests.publishers.test_django',
            ),
            ALLOWED_HOSTS = ['*'],
            ROOT_URLCONF='tests.publishers.test_django',
            MIDDLEWARE_CLASSES=(),
            TEMPLATES = [],
            LOGGING = {
                'version': 1,
                'disable_existing_loggers': True,
            },
        )

    django.setup()
    call_command('runserver',int(tests.publishers.PORT),use_reloader=False)

#_______________________________________________________________________________
class TestDjango(tests.publishers.TestPublisher):
    
    def setUp(self):
        global LM
        self.t = threading.Thread(target=_serve , daemon=True)
        self.t.start()
        time.sleep(0.2)
        self.LM = LM

# ______________________________________________________________________________
if __name__=='__main__':
    unittest.main(argv=['-v'])




"""
This module provides :py:mod:`aiohttp` routes
exposing a :py:class:`livemetrics.LiveMetrics` object.

A minimal application would be:

.. code-block:: python

    import livemetrics
    import livemetrics.publishers.aiohttp
    
    from aiohttp import web
    routes = web.RouteTableDef()

    LM = livemetrics.LiveMetrics(version,about,is_healthy,is_ready)

    @routes.get("/sample")
    @LM.timer('sample','ok','error')
    async def sample(request):
        return web.Response(status=200)

    app = web.Application()
    app.add_routes(livemetrics.publishers.aiohttp.routes(LM))   # register the routes to publish the metrics
    app.add_routes(routes)                                      # register the routes of the application
    web.run_app(ap,host,port)                                   # run the application


"""

import json

import aiohttp
from aiohttp import web

class Handler:
    def __init__(self,LM):
        self.LM = LM

    async def about(self,request):
        data = self.LM.about
        return web.Response(status=200,
            headers={"Access-Control-Allow-Origin": "*"},
            text=data)

    async def is_healthy(self,request):
        status = self.LM.is_healthy()
        if status:
            return web.Response(status=200,
                headers={"Access-Control-Allow-Origin": "*"},
                text='OK')
        else:
            return web.Response(status=500,
                headers={"Access-Control-Allow-Origin": "*"},
                text='KO')

    async def is_ready(self,request):
        status = self.LM.is_ready()
        if status:
            return web.Response(status=200,
                headers={"Access-Control-Allow-Origin": "*"},
                text='OK')
        else:
            return web.Response(status=500,
                headers={"Access-Control-Allow-Origin": "*"},
                text='KO')

    async def version(self,request):
        data = self.LM.version
        return web.Response(status=200,
            content_type='application/json',
            headers={"Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_meters3(self,request):
        event = request.match_info['event']
        result = request.match_info['result']
        metric = request.match_info['metric']
        data = self.LM.get_metrics(event,result,metric)
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_meters2(self,request):
        event = request.match_info['event']
        result = request.match_info['result']
        data = self.LM.get_metrics(event,result)
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_meters1(self,request):
        event = request.match_info['event']
        data = self.LM.get_metrics(event)
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_meters0(self,request):
        data = self.LM.get_metrics()
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_gauges2(self,request):
        object = request.match_info['object']
        metric = request.match_info['metric']
        data = self.LM.get_gauges(object,metric)
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_gauges1(self,request):
        object = request.match_info['object']
        data = self.LM.get_gauges(object)
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_gauges0(self,request):
        data = self.LM.get_gauges()
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    def _get_histograms(self,request,event,metric):
        msg = None
        params = {}
        params.update(request.query)
        if not isinstance(params.setdefault('percentiles',[0.05, 0.25, 0.5, 0.75, 0.95]),list):
            params['percentiles'] = [params['percentiles']]
        if not isinstance(params.setdefault('scale',[10]),list):
            params['scale'] = [params['scale']]

        if len(params)!=2:
            keys = ", ".join(sorted(list(set(params.keys()) - {'percentiles', 'scale'})))
            msg="Invalid query parameters: " + keys
        if not metric in [None,'count','min','max','mean','stddev','quantiles','distribution']:
            msg = "Unknown metric " + metric
        if msg:
            return web.Response(status=400,body=msg)

        # Make sure the types are correct
        params['percentiles'] = [float(x) for x in params['percentiles']]
        params['scale'] = int(params['scale'][0])

        try:
            data = self.LM.get_histograms(event,metric,**params)
        except Exception as exc:
            msg = str(exc)
            return web.Response(status=500,body=msg)
        data = json.dumps(data)
        return web.Response(status=200,
            content_type='application/json',
            headers={"Content-Length": str(len(data)), "Access-Control-Allow-Origin": "*"},
            body=data)

    async def get_histograms2(self,request):
        event = request.match_info['event']
        metric = request.match_info['metric']
        return self._get_histograms(request,event,metric)

    async def get_histograms1(self,request):
        event = request.match_info['event']
        return self._get_histograms(request,event,None)

    async def get_histograms0(self,request):
        return self._get_histograms(request,None,None)

def routes(LM):
    """
    Return a list of routes to be registered in the :py:mod:`aiohttp` application.

    *LM*: a :py:class:`livemetrics.LiveMetrics` object
    """
    handler = Handler(LM)
    return [
        web.get('/monitoring/v1/about', handler.about),
        web.get('/monitoring/v1/is_healthy', handler.is_healthy),
        web.get('/monitoring/v1/is_ready', handler.is_ready),
        web.get('/monitoring/v1/version', handler.version),

        web.get('/monitoring/v1/metrics/meters/{event}/{result}/{metric}', handler.get_meters3),
        web.get('/monitoring/v1/metrics/meters/{event}/{result}', handler.get_meters2),
        web.get('/monitoring/v1/metrics/meters/{event}', handler.get_meters1),
        web.get('/monitoring/v1/metrics/meters', handler.get_meters0),

        web.get('/monitoring/v1/metrics/gauges/{object}/{metric}', handler.get_gauges2),
        web.get('/monitoring/v1/metrics/gauges/{object}', handler.get_gauges1),
        web.get('/monitoring/v1/metrics/gauges', handler.get_gauges0),

        web.get('/monitoring/v1/metrics/histograms/{event}/{metric}', handler.get_histograms2),
        web.get('/monitoring/v1/metrics/histograms/{event}', handler.get_histograms1),
        web.get('/monitoring/v1/metrics/histograms', handler.get_histograms0),
    ]





import unittest
import time
import random
import threading
import json
import requests
import http.server

import livemetrics
import livemetrics.publishers.http

import tests.publishers

LM = livemetrics.LiveMetrics('{"version":"1.0"}',"Test server",True, True)

# Sample of gauges
LM.gauge('fixed',10)
LM.gauge('test',10)
LM.gauge('call',lambda: random.randrange(100,1000))

#______________________________________________________________________________
def ok(ret):
    return str(ret.status)

def error(exc):
    return str(exc)
    
def exc(ret):
    raise Exception("Bad status")

COUNT = 0

#______________________________________________________________________________
class TestHTTPRequestHandler(livemetrics.publishers.http.HTTPRequestHandler):

    @LM.timer("decorator_async","ok","error")
    @LM.timer("decorator_async_callable",ok,error)
    @LM.timer("decorator_async_histo_only",None,None)
    @LM.timer("decorator_async_callable_exc",exc,exc)
    def _doit(self):
        time.sleep(random.random()/10.)
        LM.mark('test',"ok")
        LM.gauge('test',random.randrange(10,10000))
        LM.histogram("histo",random.random()*20)
        global COUNT
        COUNT += 1
        if COUNT % 3 ==0:
            raise Exception("false-alert")

    def do_GET(self):
        if self.path=='/test':
            try:
                self._doit()
                self.send_response(200)
                self.end_headers()
            except:
                self.send_response(500)
                self.end_headers()
        else:
            return super().do_GET()

    def log_message(self, format, *args):
        pass
        
def _serve():
    global LM
    tests.publishers.PORT = '8765'
    server_address = (tests.publishers.IP, int(tests.publishers.PORT))
    httpd = http.server.HTTPServer(server_address, lambda r,a,s: TestHTTPRequestHandler(LM,r,a,s))
    global HTTPD
    HTTPD = httpd
    httpd.allow_reuse_address = True
    httpd.serve_forever()

#_______________________________________________________________________________
class TestHttp(tests.publishers.TestPublisher):
    
    def setUp(self):
        t = threading.Thread(target=_serve, daemon=False)
        t.start()
        time.sleep(0.2)
        self.LM = LM

    def tearDown(self):
        # Clean up to release the socket address for the other tests
        global HTTPD
        HTTPD.shutdown()

# ______________________________________________________________________________
if __name__=='__main__':
    unittest.main(argv=['-v'])




# Inspired by https://metrics.dropwizard.io/4.0.0/

import math
import threading
import datetime
import random
import bisect

#______________________________________________________________________________
class Gauge(object):
    """
    A simple counter, registering a value and statistics (min and max) on this value.

    >>> g = Gauge(10)
    >>> g.mark(12)
    >>> g.min,g.count,g.max
    (10, 12, 12)

    >>> g.mark(8)
    >>> g.min,g.count,g.max
    (8, 8, 12)

    >>> g = Gauge()
    >>> g.mark(12)
    >>> g.min,g.count,g.max
    (12, 12, 12)

    >>> g.mark(8)
    >>> g.min,g.count,g.max
    (8, 8, 12)

    >>> g.mark(20)
    >>> g.min,g.count,g.max
    (8, 20, 20)

    >>> g.mark(15)
    >>> g.min,g.count,g.max
    (8, 15, 20)

    The alias ``value`` can also be used:

    >>> g.min,g.value,g.max
    (8, 15, 20)

    Any time, the value can be provided as a callable:

    >>> import random
    >>> g.mark(lambda: random.randrange(21,50))
    >>> g.count>20
    True

    The value can also be a callable from the start

    >>> g = Gauge(lambda: random.randrange(1,20))
    >>> g.mark(10)
    >>> g.min>=0,g.max<20
    (True, True)

    """

    def __init__(self,value=None):
        self.lock = threading.RLock()
        self._callable = None
        if callable(value):
            self._callable = value
            value = self._callable()
        self._value = value
        self._min = value
        self._max = value

    def mark(self,value):
        """
        Register a new value in the gauge and update the statistics.
        """
        with self.lock:
            if callable(value):
                self._callable = value
                value = self._callable()
            self._value = value
            self._min = min(self._min,value) if self._min else value
            self._max = max(self._max,value) if self._max else value

    @property
    def count(self):
        """
        Return the current value of the gauge.
        """
        if self._callable:
            self.mark(self._callable())
        return self._value

    @property
    def value(self):
        """
        Alias on :py:meth:`count`.
        """
        return self.count

    @property
    def min(self):
        """
        Return the minimum value ever registered in this gauge.
        """
        return self._min

    @property
    def max(self):
        """
        Return the maximum value ever registered in this gauge.
        """
        return self._max

    def to_dict(self):
        return dict(
            min=self.min,
            max=self.max,
            count=self.count
        )

#______________________________________________________________________________
class EWMA(object):
    """
    An exponentially-weighted moving average.
    See http://en.wikipedia.org/wiki/Moving_average#Exponential_moving_average

    Usage:

    >>> e = EWMA(1)
    >>> e.update()
    >>> import time
    >>> time.sleep(1)
    >>> e.update()
    >>> e._tick()
    >>> e.rate  #doctest: +ELLIPSIS
    0.39999...
    >>> e._tick()
    >>> e.rate  #doctest: +ELLIPSIS
    0.368017...
    """

    def __init__(self,nb_minutes=1,interval=5):
        self.initialized = False
        self._rate = 0.0

        self.uncounted = []
        self.alpha = 0.0
        self.interval = float(interval)*1000000.    # convert to microseconds
        self.alpha = 1.0 - math.exp(-interval / 60.0 / nb_minutes)

    def update(self,n=1):
        """
        Indicate an event has happened. *n* specifies the number of events.
        """
        self.uncounted.append(n)

    def _tick(self):
        # Mark the passage of time and decay the current rate accordingly.
        count = sum(self.uncounted)
        self.uncounted = []
        instantRate = count/self.interval
        if self.initialized:
            self._rate = self._rate + (self.alpha*(instantRate-self._rate))
        else:
            self._rate = instantRate
            self.initialized = True

    @property
    def rate(self):
        return self._rate*1000000.

#______________________________________________________________________________
class Meter(object):
    """
    An object measuring the rates of occurence of an event.
    Each event is reported with a call to :py:meth:`mark`. The rates are decayed as time
    goes by.

    For testing, we will reduce the tick interval managing the decaying of values.

    >>> Meter.TICK_INTERVAL = 0.1

    Then we create a new object:

    >>> mtr = Meter()

    At start, it is empty:

    >>> mtr.mean
    0.0

    Then if we mark an event:

    >>> mtr.mark()

    Mean is still 0 because there is no enough time difference to calculate it.

    >>> mtr.mean
    0.0

    Let's move forward in time:

    >>> import time
    >>> time.sleep(0.1)

    and report a second event:

    >>> mtr.mark()
    >>> mtr.mean>17 and mtr.mean<25
    True
    >>> mtr.count
    2
    >>> mtr.mark()
    >>> time.sleep(0.5)
    >>> 10.0-mtr.rate1<0.1,10.0-mtr.rate5<0.1,10.0-mtr.rate15<0.1
    (True, True, True)

    Rates decreases as time goes:

    >>> time.sleep(1.0)
    >>> 9.8-mtr.rate1<0.1,9.9-mtr.rate5<0.1,10.0-mtr.rate15<0.1
    (True, True, True)

    We marked 3 times in 2.1 seconds

    >>> round(mtr.mean,1)
    1.9
    """
    TICK_INTERVAL = 5

    def __init__(self):
        self.lock = threading.RLock()
        self.ewma1 = EWMA(1,self.TICK_INTERVAL)
        self.ewma5 = EWMA(5,self.TICK_INTERVAL)
        self.ewma15 = EWMA(15,self.TICK_INTERVAL)
        self._count = 0
        self.start = datetime.datetime.now()
        self.last_tick = datetime.datetime.now()

    def mark(self):
        """
        Indicate an event has just occurred.
        """
        with self.lock:
            self._tick_if_necessary()
            self._count += 1
            self.ewma1.update()
            self.ewma5.update()
            self.ewma15.update()

    def _tick_if_necessary(self):
        new_tick = datetime.datetime.now()
        age = new_tick - self.last_tick
        while age.seconds*1000.+age.microseconds/1000.>self.TICK_INTERVAL*1000.:
            self.last_tick += datetime.timedelta(seconds=self.TICK_INTERVAL)
            age = new_tick - self.last_tick
            self.ewma1._tick()
            self.ewma5._tick()
            self.ewma15._tick()

    @property
    def mean(self):
        """
        The number of events divided by the time the application is running.
        This gives a very rough estimation of the average number of events
        per seconds.
        """
        if self._count==0:
            return 0.0
        period = datetime.datetime.now()-self.start
        period = period.seconds*1.0 + period.microseconds/1000000.
        if period<=0.1:
            return 0.0            
        return self._count/period

    @property
    def count(self):
        """
        Return the number of events reported to this object since it was created.
        """
        return self._count

    @property
    def rate1(self):
        """
        Return the 1-minute moving average rate.
        """
        with self.lock:
            self._tick_if_necessary()
            return self.ewma1.rate

    @property
    def rate5(self):
        """
        Return the 5-minute moving average rate.
        """
        with self.lock:
            self._tick_if_necessary()
            return self.ewma5.rate

    @property
    def rate15(self):
        """
        Return the 15-minute moving average rate.
        """
        with self.lock:
            self._tick_if_necessary()
            return self.ewma15.rate

    def to_dict(self):
        return dict(
            mean=self.mean,
            count=self.count,
            rate1=self.rate1,
            rate5=self.rate5,
            rate15=self.rate15
        )

#______________________________________________________________________________
class WeightedSample(object):
    def __init__(self,value,weight):
        self.value = value
        self.weight = weight

    def __repr__(self):
        return "value={}; weight={}".format(self.value,self.weight)

class WeightedSnapshot(object):
    """
    A (weighted) snapshot, i.e. a copy of a set of values that have a weight.
    The weight is used to calculate the quantiles.

    This class is used by :py:class:`Histogram` and returned by :py:attr:`Histogram.snapshot`

    Example:

    >>> snapshot = WeightedSnapshot( {i:WeightedSample(i,i) for i in range(10)})

    Access the statistics of the values:

    >>> print((snapshot.min,snapshot.max,round(snapshot.mean,1),round(snapshot.stddev,1)))
    (0, 9, 6.3, 2.2)

    >>> print(snapshot.get_value(0.5))
    7.5

    >>> print(snapshot.get_value(-1))
    Traceback (most recent call last):
    ...
    Exception: argument -1 is not in [0..1]

    >>> print(snapshot.get_value(2))
    Traceback (most recent call last):
    ...
    Exception: argument 2 is not in [0..1]

    >>> print(snapshot.get_distribution(5))
    [2, 2, 2, 2, 2]

    >>> print(snapshot.size)
    10

    The snapshot can also be empty:

    >>> snapshot = WeightedSnapshot( {})
    >>> print((snapshot.min,snapshot.max,round(snapshot.mean,1),round(snapshot.stddev,1)))
    (0, 0, 0, 0)

    >>> print(snapshot.get_value(0.5))
    0.0

    """
    def __init__(self,values):
        copy = [v for v in values.values()]
        copy.sort(key=lambda x: x.value)

        sum_weight = sum([x.weight for x in copy])
        self.values = [x.value for x in copy]
        self.norm_weights = [x.weight/sum_weight if sum_weight!=0 else 0 for x in copy]
        self.quantiles = [0.0]*len(copy)
        for i in range(1,len(copy)):
            self.quantiles[i] = self.quantiles[i-1] + self.norm_weights[i-1]

    def get_value(self,quantile):
        """
        Access the value (the quantile) for a given percentile, i.e. the value that defines the
        boundary between the two segment [O,percentile] and  [percentile,1].

        The percentile must be between 0 and 1.

        """
        if quantile < 0.0 or quantile > 1.0 or quantile is None:
            raise Exception("argument %s is not in [0..1]" % quantile)

        if len(self.values)==0:
            return 0.0

        posx = bisect.bisect_left(self.quantiles,quantile)
        if posx<0:
            posx = ((-posx) - 1) - 1

        if posx<1:
            return self.values[0]

        if posx >= len(self.values):
            return self.values[-1]

        return (self.values[posx]+self.values[posx-1])/2

    def get_distribution(self,range_value=10):
        """
        Return the distribution with the request resolution (i.e. number of values)
        """
        H = []
        posy = 0
        step = (float(self.max)-float(self.min))/float(range_value)
        x = self.min + step
        while x<self.max:
            nposy = bisect.bisect_left(self.values,x)
            H.append(nposy-posy)
            posy = nposy
            x += step
        nposy = len(self.values)
        H.append(nposy-posy)
        return H

    @property
    def size(self):
        """
        Number of values in this snapshot.
        """
        return len(self.values)

    @property
    def max(self):
        """
        Maximum value in this snapshot.
        """
        if len(self.values)==0:
            return 0
        return self.values[-1]

    @property
    def min(self):
        """
        Minimum value in this snapshot.
        """
        if len(self.values)==0:
            return 0
        return self.values[0]

    @property
    def mean(self):
        """
        Weighted average of the values in this snapshot.
        """
        if len(self.values)==0:
            return 0
        return sum( [self.values[i]*self.norm_weights[i] for i in range(len(self.values)) ] )

    @property
    def stddev(self):
        """
        Standard deviation of the weighted values in this snapshot.
        """
        if len(self.values)<=1:
            return 0
        mean = self.mean
        variance = 0.0
        for i in range(len(self.values)):
            diff = self.values[i] - mean
            variance += self.norm_weights[i]*diff*diff
        return math.sqrt(variance)

class Reservoir(object):
    """
    A reservoir holds the data of an :py:class:`Histogram`.
    It records values and decays them as time passes.
    It keeps only a statistical representative set of values.
    """
    TICK_INTERVAL = 1.0*60*60     # 1 hour expressed in seconds

    def __init__(self):
        self.values = {}
        self.alpha = 0.015
        self._size = 1028
        self.count = 0
        self.start = datetime.datetime.now()
        self.last_tick = datetime.datetime.now()

    @property
    def size(self):
        return min(self._size,self.count)

    def update(self,value):
        self._tick_if_necessary()
        period = datetime.datetime.now()-self.start
        period = period.seconds*1.0 + period.microseconds/1000000.
        item_weight = math.exp(self.alpha*period)
        sample = WeightedSample(value, item_weight)
        priority = item_weight / random.random()
        self.count += 1
        if self.count<=self.size:
            self.values[priority] = sample
        else:
            first = min(self.values.keys())
            if first<priority:
                if not priority in self.values:
                    self.values[priority] = sample
                    del self.values[first]
                else:
                    self.values[priority] = sample

    def _tick_if_necessary(self):
        new_tick = datetime.datetime.now()
        age = new_tick - self.last_tick
        while age.seconds+age.microseconds/1000000.>self.TICK_INTERVAL:
            self.last_tick += datetime.timedelta(seconds=self.TICK_INTERVAL)
            age = new_tick - self.last_tick
            self.rescale()

    def rescale(self):
        period = self.TICK_INTERVAL
        scaling_factor = math.exp(-self.alpha * period)
        if math.isclose(scaling_factor,0.0):
            self.values = {}
        else:
            values = {}
            for k,v in self.values.items():
                new_sample = WeightedSample(v.value, v.weight * scaling_factor)
                if math.isclose(new_sample.weight,0.0):
                    continue
                values[k*scaling_factor] = new_sample
            self.values = values
        self.count = len(self.values)
        self.start += datetime.timedelta(seconds=self.TICK_INTERVAL)


    @property
    def snapshot(self):
        self._tick_if_necessary()
        return WeightedSnapshot(self.values)


class Histogram(object):
    """
    An histogram represents the distribution of a set of values.
    Values are exponentially decayed so that more recent values have more weight than old values.

    For testing, we will reduce the tick interval managing the decaying of values.

    >>> Reservoir.TICK_INTERVAL = 0.1

    Usage:

    >>> his = Histogram()

    A new object has no value yet:

    >>> his.count
    0

    For the test, we 
    Let's add 100 sequential values:

    >>> for i in range(100):
    ...    his.update(i+1)

    The ``snapshot`` object gives access to some statistics:

    >>> round(his.snapshot.mean,1)
    50.5
    >>> round(his.snapshot.stddev,1)
    28.9

    And to the value for a percentile:

    >>> round(his.snapshot.get_value(0.25),1)
    26.5

    >>> his.snapshot.get_distribution()
    [10, 10, 10, 10, 10, 10, 10, 10, 10, 10]

    >>> his.snapshot.get_distribution(20)
    [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]

    If we wait a little and inject new values:

    >>> import time
    >>> time.sleep(1.0)
    >>> for i in range(100):
    ...    his.update(i*2)

    >>> his.snapshot.get_distribution(20)
    [14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5]
    >>> round(his.snapshot.get_value(0.25),1)
    34.0

    >>> time.sleep(1.0)
    >>> for i in range(100):
    ...    his.update(i*2)
    >>> his.snapshot.get_distribution(20)
    [19, 20, 20, 20, 20, 20, 20, 20, 20, 20, 11, 10, 10, 10, 10, 10, 10, 10, 10, 10]
    >>> round(his.snapshot.get_value(0.25),1)
    38.0

    The size of the reservoir is by default 1024. When the number of values is over this
    limit, values are randomly removed, older values first.

    >>> his50 = Histogram()
    >>> his50._reservoir._size = 50
    >>> for i in range(100):
    ...    his50.update(i+1)

    >>> his.snapshot.mean!=his50.snapshot.mean
    True
    >>> his.snapshot.stddev!=his50.snapshot.stddev
    True

    """

    def __init__(self):
        self.lock = threading.RLock()
        self._count = 0
        self._reservoir = Reservoir()

    def update(self,value):
        """
        Register a new value in the histogram.
        """
        with self.lock:
            self._count += 1
            self._reservoir.update(value)

    @property
    def count(self):
        """
        Get the number of values recorded by this histogram.
        """
        return self._count

    @property
    def snapshot(self):
        """
        Get a :py:class:`WeightedSnapshot` for this histogram. The snapshot will then give
        access to quantiles and to the distribution of values.
        """
        return self._reservoir.snapshot

    def to_dict(self):
        snapshot = self.snapshot
        return dict(
            count=self.count,
            min=snapshot.min,
            max=snapshot.max,
            mean=snapshot.mean,
            stddev=snapshot.stddev,
        )




import unittest
import time
import random
import threading
import json
import requests
import asyncio

import aiohttp
from aiohttp import web
routes = web.RouteTableDef()

import livemetrics
import livemetrics.publishers.aiohttp

import tests.publishers

LM = livemetrics.LiveMetrics('{"version":"1.0"}',"Test server",True, True)

# Sample of gauges
LM.gauge('fixed',10)
LM.gauge('test',10)
LM.gauge('call',lambda: random.randrange(100,1000))

#______________________________________________________________________________
def ok(ret):
    return str(ret.status)

def error(exc):
    return str(exc)

def exc(ret):
    raise Exception("Bad status")

COUNT = 0

@routes.get('/test')
@LM.timer("decorator_async","ok","error")
@LM.timer("decorator_async_callable",ok,error)
@LM.timer("decorator_async_histo_only",None,None)
@LM.timer("decorator_async_callable_exc",exc,exc)
async def get(request):
    time.sleep(random.random()/10.)
    LM.mark('test',"ok")
    LM.gauge('test',random.randrange(10,10000))
    LM.histogram("histo",random.random()*20)
    global COUNT
    COUNT += 1
    if COUNT % 3 ==0:
        raise Exception("false-alert")
    return web.Response(status=random.choice([200,500]))

async def runner():
    app = web.Application()
    app.add_routes(livemetrics.publishers.aiohttp.routes(LM))
    app.add_routes(routes)
    runner = web.AppRunner(app)
    await runner.setup()

    tests.publishers.PORT = '8766'
    site = web.TCPSite(runner, tests.publishers.IP, int(tests.publishers.PORT),reuse_address=True)
    await site.start()

def run_server(handler):
    global LOOP
    loop = asyncio.new_event_loop()
    LOOP = loop
    asyncio.set_event_loop(loop)
    loop.run_until_complete(handler)

    try:
        loop.run_forever()
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()

#_______________________________________________________________________________
class TestAioHttp(tests.publishers.TestPublisher):
    
    def setUp(self):
        self.t = threading.Thread(target=run_server,args=(runner(),) , daemon=False)
        self.t.start()
        time.sleep(0.2)
        self.LM = LM

    def tearDown(self):
        # Clean up to release the socket address for the other tests
        global LOOP
        LOOP.stop()
            
# ______________________________________________________________________________
if __name__=='__main__':
    unittest.main(argv=['-v'])




import unittest
import time
import random
import threading

import requests

import livemetrics
import livemetrics.publishers.flask

from flask import Flask
from flask import request

app = Flask(__name__)

import tests.publishers

LM = livemetrics.LiveMetrics('{"version":"1.0"}',"Test server",True, True)

# Sample of gauges
LM.gauge('fixed',10)
LM.gauge('test',10)
LM.gauge('call',lambda: random.randrange(100,1000))

#______________________________________________________________________________
def ok(ret):
    return str(ret.status)

def error(exc):
    return str(exc)

def exc(ret):
    raise Exception("Bad status")

COUNT = 0

@app.route('/test')
@LM.timer("decorator_async","ok","error")
@LM.timer("decorator_async_callable",ok,error)
@LM.timer("decorator_async_histo_only",None,None)
@LM.timer("decorator_async_callable_exc",exc,exc)
def get():
    time.sleep(random.random()/10.)
    LM.mark('test',"ok")
    LM.gauge('test',random.randrange(10,10000))
    LM.histogram("histo",random.random()*20)
    global COUNT
    COUNT += 1
    if COUNT % 3 ==0:
        raise Exception("false-alert")
    return "ok" #web.Response(status=random.choice([200,500]))

def shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()

@app.route('/shutdown', methods=['POST'])
def shutdown():
    shutdown_server()
    return 'Server shutting down...'

def _serve():
    global LM
    tests.publishers.PORT = '8767'
    global app
    app.register_blueprint(livemetrics.publishers.flask.blueprint(LM))
    app.run(host=tests.publishers.IP, port=int(tests.publishers.PORT))

#_______________________________________________________________________________
class TestFlask(tests.publishers.TestPublisher):
    
    def setUp(self):
        self.t = threading.Thread(target=_serve , daemon=False)
        self.t.start()
        time.sleep(0.2)
        self.LM = LM

    def tearDown(self):
        # Clean up to release the socket address for the other tests
        requests.post('http://'+tests.publishers.IP+':'+tests.publishers.PORT+'/shutdown')

            
# ______________________________________________________________________________
if __name__=='__main__':
    unittest.main(argv=['-v'])




import unittest
import doctest
import livemetrics.metrics

# Used by: python setup.py test
def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(livemetrics.metrics))
    return tests

load_tests.__test__ = False

# ______________________________________________________________________________
if __name__=='__main__':
    doctest.testmod(livemetrics.metrics)



import time
import unittest
import requests
import json
import math
import statistics

from livemetrics.metrics import *

#_______________________________________________________________________________
class TestMetrics(unittest.TestCase):
    
    def assertEqualFloat(self,a,b,msg=None):
        # We want less than 1.5% of delta between the 2 values
        if math.fabs(a)<0.00001 and math.fabs(b)<0.00001:
            # both are null
            return
        if math.fabs(b)>=0.00001:
            r = a/b
        else:
            r = b/a
        if r<0.985 or r>1.015:
            self.fail(self._formatMessage(msg, "{} != {} (error: {}%)".format(a,b,round(r,3)*100)))

    def assertEqualQuantile(self,a,b,msg=None):
        if math.fabs(b-a)>1.0:
            self.fail(self._formatMessage(msg, "{} != {}".format(a,b)))

    def test_simple(self):
        self.maxDiff = None
        his = Histogram()
        l = [10, 10, 10, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
        for x in l:
            his.update(x)
        self.assertEqual(l,[x.value for x in his._reservoir.values.values()])
        self.assertEqualFloat(statistics.mean(l),his.snapshot.mean)
        self.assertEqualFloat(statistics.pstdev(l),his.snapshot.stddev)
        self.assertEqualFloat(statistics.median(l),his.snapshot.get_value(0.5))

    def test_histograms(self):
        self.maxDiff = None
        his = Histogram()
        l = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,25,26,27,28,29,30,31,32,33,34,35,36,37,38,
            39,40,41,42,43,44,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,75,76,77,78,79,
            80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119]
        for x in l:
            his.update(x)
        self.assertEqual(l,[x.value for x in his._reservoir.values.values()])
        self.assertEqualFloat(statistics.mean(l),his.snapshot.mean)
        self.assertEqualFloat(statistics.pstdev(l),his.snapshot.stddev)
        self.assertEqualQuantile(statistics.median(l),his.snapshot.get_value(0.5))

        self.assertEqual(0,his.snapshot.min)
        self.assertEqual(119,his.snapshot.max)
        self.assertEqual(100,his.snapshot.size)
        #self.assertEqual(61.57,round(his.snapshot.mean,2))
        self.assertEqualFloat(35.93,his.snapshot.stddev)
        self.assertEqualQuantile(61.5,round(his.snapshot.get_value(0.5),1))


# ______________________________________________________________________________
if __name__=='__main__':
    unittest.main(argv=['-v'])




#!/usr/bin/env python

import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

about = {}
with open('livemetrics/__init__.py', 'r') as f:
    exec(f.read(), about)

setuptools.setup(
    name = 'livemetrics',
    version = about['__version__'],
    author = about['__author__'],
    author_email = "olivier.heurtier@idemia.com",
    license = about['__license__'],
    description = 'LiveMetrics collector and publisher for Python applications',
    long_description = long_description,
    url="https://github.com/idemia/python-livemetrics",
    packages = ['livemetrics','livemetrics.publishers','livemetrics.dashboard'],
    package_data={'livemetrics.dashboard': ['dashboard.html']},
    test_suite = 'tests',
    extras_require = {
        'aiohttp': [
            'aiohttp >= 3.5',
        ],
        'django': [
            'django >= 2.0',
        ],
        'flask': [
            'flask >= 1.0.0',
        ],
        'dashboard': [
            'jinja2 >=2.10.0',
            'PyYAML >= 5.0.0',
        ],
    },
    tests_require = [
        'requests',
        'aiohttp >= 3.5',
        'django >= 2.0',
        'flask >= 1.0.0',
        'jinja2 >=2.10.0',
        'PyYAML >= 5.0.0',
    ],
    entry_points = {
        'console_scripts': ['livemetrics-dashboard=livemetrics.dashboard.dashboard:main'],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: CeCILL-C Free Software License Agreement (CECILL-C)",
        "Operating System :: OS Independent",
    ],
)




import time
import requests
import json
import random

def inject():
    # Send some events
    for y in range(5):
        time.sleep(0.5)

        for x in range(20):
            requests.get('http://localhost:7070/test/'+str(y*25+x))

    for y in range(10,15):
        for x in range(30):
            requests.get('http://localhost:7070/test/'+str(y*25+x))

    for x in range(100):
        requests.get('http://localhost:7070/test/'+str(random.randrange(300)))


# ______________________________________________________________________________
if __name__=='__main__':
    inject()





"""
This module implements a utility to collect (efficiently) business
metrics showing the activity of an application.

For testing, we will reduce the tick interval

>>> import livemetrics.metrics
>>> livemetrics.metrics.Meter.TICK_INTERVAL = 0.1

First you need to create an object and to send an event:

>>> lm = LiveMetrics("version","about",lambda: True,lambda: True)
>>> lm.mark('start','ok')
>>> lm.mark('start','error')

This is enough to get a first set of metrics:

>>> import pprint
>>> pprint.pprint(lm.get_metrics()) #doctest: +ELLIPSIS
{'start': {'error': {'count': 1,
...
                     'rate1': 0.0,
                     'rate15': 0.0,
                     'rate5': 0.0},
           'ok': {'count': 1,
...
                  'rate1': 0.0,
                  'rate15': 0.0,
                  'rate5': 0.0}}}

We can now emit more events:

>>> import time
>>> time.sleep(0.02)
>>> lm.mark('start','ok')
>>> time.sleep(0.03)
>>> lm.mark('start','ok')
>>> lm.mark('start','error')
>>> time.sleep(0.2)
>>> round(lm.get_metrics()['start']['ok']['rate1'],2)
29.95
>>> round(lm.get_metrics()['start']['error']['rate1'],2)
19.97
>>> m = lm.get_metrics()
>>> m['start']['error']['rate1']<m['start']['error']['rate5']
True
>>> m['start']['error']['rate5']<m['start']['error']['rate15']
True
>>> m['start']['error']['rate1']<m['start']['ok']['rate1']
True
>>> m['start']['ok']['rate1']>29.0
True
>>> m['start']['error']['rate1']>19.0
True

For gauges, here is an example:

>>> lm = LiveMetrics("version","about", lambda: True, memory_and_cpu=False)
>>> lm.gauge('G',10)
>>> lm.gauge('G',15)
>>> lm.gauge('G',20)
>>> lm.gauge('G',12)

>>> print(lm.get_gauges())
{'G': {'min': 10, 'max': 20, 'count': 12}}

>>> print(lm.get_gauges('G'))
{'min': 10, 'max': 20, 'count': 12}

>>> print(lm.get_gauges('G','max'))
20

This module also provides a decorator to encapsulate a meter and a histogram.
Example:

>>> lm = LiveMetrics("version","about",lambda: True)
>>> @lm.timer('decorated','OK','EXC')
... def for_test(value,exc=False):
...     if exc: raise Exception("Exception")
...     time.sleep(0.1)
...     return value

Then, every time the function is called, new metrics are collected:

>>> for_test(1)
1
>>> for_test(2)
2
>>> for_test(3,True)
Traceback (most recent call last):
...
Exception: Exception

Metrics can be accessed using :py:meth:`LiveMetrics.get_metrics`:

>>> pprint.pprint(lm.get_metrics()) #doctest: +ELLIPSIS
{'decorated': {'EXC': {'count': 1,
...
                       'rate5': 0.0},
               'OK': {'count': 2,
...
                      'rate5': 10.0}}}

>>> pprint.pprint(lm.get_metrics('decorated')) #doctest: +ELLIPSIS
{'EXC': {'count': 1, 'mean': 0.0, 'rate1': 0.0, 'rate15': 0.0, 'rate5': 0.0},
 'OK': {'count': 2,
...
        'rate5': 10.0}}

>>> M = lm.get_metrics('decorated','OK')
>>> M['count']
2
>>> round(M['rate5'],1)
10.0

>>> lm.get_metrics('decorated','OK','count')
2

Event must be specified if result and metric are defined:
>>> lm.get_metrics(None,'OK','count')
Traceback (most recent call last):
...
SyntaxError: if metric/result is specified, event must also be specified
>>> lm.get_metrics('decorated',None,'count')
Traceback (most recent call last):
...
SyntaxError: if metric is specified, result must also be specified

Or :py:meth:`LiveMetrics.get_histograms`:

>>> H = lm.get_histograms()
>>> H['decorated']['count']
3
>>> round(H['decorated']['quantiles'][0.05],1)
0.1
>>> round(H['decorated']['stddev'],2)
0.05

>>> H = lm.get_histograms('decorated')
>>> H['count']
3
>>> round(H['quantiles'][0.05],1)
0.1
>>> round(H['stddev'],2)
0.05

>>> pprint.pprint(lm.get_histograms('decorated','count')) #doctest: +ELLIPSIS
3

>>> math.isclose(lm.get_histograms('decorated','quantiles')[0.05],0.05,rel_tol=0.01)
True

>>> pprint.pprint(lm.get_histograms('decorated','distribution',scale=1)) #doctest: +ELLIPSIS
[3]

"""

import os
import time
import collections
from functools import wraps
import asyncio

from livemetrics.metrics import *

__version__ = '0.3'
__author__ = "Olivier Heurtier"
__copyright__ = "IDEMIA"
__license__ = "CeCILL-C"

if os.name=='posix':
    def get_memory():
        # Retrieve memory usage from /proc/self/statm
        try:
            with open('/proc/self/statm','r') as f:
                return int(f.read().split(' ')[0]) *  os.sysconf('SC_PAGE_SIZE')
        except:
            return 0

    __CPU = None
    __CPU_RESULT = 0
    def get_cpu():
        # Retrieve CPU usage from /proc/self/stat
        global __CPU
        global __CPU_RESULT
        try:
            with open('/proc/self/stat','r') as f:
                # 14th value is the user time
                # (note: 20th is the number of threads, could be useful)
                utime = float(f.read().split(' ')[14-1]) / os.sysconf('SC_CLK_TCK')
                now = time.time()
                if __CPU is None:
                    __CPU = (now,utime)
                    __CPU_RESULT = 0
                    return __CPU_RESULT

                # Getting 2 consecutive values very close can give bad value
                if now-__CPU[0] < 0.3:
                    return __CPU_RESULT
                result = (utime-__CPU[1])/(now-__CPU[0])
                __CPU = (now,utime)
                __CPU_RESULT = int(result*100)
                return __CPU_RESULT
        except:
            return 0
                
#______________________________________________________________________________
class LiveMetrics(object):
    """
    An object to be used as a singleton to aggregate all live metrics of an application.

    """

    def __init__(self,version,about,is_healthy,is_ready=None,memory_and_cpu=True):
        """
        Contructor.

        *version*: a string giving the version of the application

        *about*: a string describing the application

        *is_healthy*: a function called to know if the application is healthy or not.
        Function must have no parameter and return a boolean.

        *is_ready*: a function called to know if the application is ready to process requests or not.
        Function must have no parameter and return a boolean.
        If None is provided, *is_healthy* is used.
        
        *memory_and_cpu*: a flag to activate gauges to report the memory and CPU usage (on Linux only)
        Default is True.
        
        """
        self.version = version
        self.about = about
        self.is_healthy = is_healthy
        if not callable(self.is_healthy):
            self.is_healthy = lambda: is_healthy

        self.is_ready = is_ready
        if self.is_ready is None:
            self.is_ready = self.is_healthy
        if not callable(self.is_ready):
            self.is_ready = lambda: is_ready

        # Init the structure to receive the metrics
        self._meters = collections.defaultdict( lambda: collections.defaultdict(Meter) )
        self._gauges = collections.defaultdict( Gauge )
        self._histograms = collections.defaultdict( Histogram )

        if memory_and_cpu and os.name=='posix':
            self.gauge('memory',get_memory)
            self.gauge('cpu',get_cpu)

    def mark(self,event,result):
        """
        Mark the execution of an event **event** with the **result**.
        """
        self._meters[event][result].mark()

    def gauge(self,name,value):
        """
        Register a new gauge for this **name**. If **value** is a callable,
        it will be called everytime the gauge value is accessed.
        """
        self._gauges[name].mark(value)

    def histogram(self,name,value):
        """
        Register a new value in a histogram.
        """
        self._histograms[name].update(value)

    def timer(self,event,ok,error):
        """
        Decorator to automate meter and histogram on one event.

        *event*: the name of the event

        *ok*: value when the function decorated terminates successfully

        *error*: value when the function decorated terminates with an exception

        For each call of the decorated function, a call to :py:meth:`mark` is done for
        the *event* and result (*ok* or *error*), and a call to :py:meth:`histogram` is
        made to measure the processing time.

        *ok* and *error* can be callables that return a string compatible with json dictionary key

        .. warning::

            If applied on an async function decorated with aiohttp :py:meth:`web.RouteTableDef`, it must be
            placed between the aiohttp annotation and the async function. Otherwise the aiohttp route
            will be decorated and not the actual function.
        """

        def _f(f,event=event,ok=ok,error=error):
            @wraps(f)
            async def __asyncf(*args,**kw):
                S = time.time()
                try:
                    ret = await f(*args,**kw)
                    try:
                        if callable(ok):
                            self.mark(event,ok(ret))
                        elif ok is not None:
                            self.mark(event,ok)
                    except:
                        pass
                    E = time.time()
                    self.histogram(event,E-S)
                    return ret
                except Exception as exc:
                    try:
                        if callable(error):
                            self.mark(event,error(exc))
                        elif error is not None:
                            self.mark(event,error)
                    except:
                        pass   
                    E = time.time()
                    self.histogram(event,E-S)
                    raise

            @wraps(f)
            def __f(*args,**kw):
                S = time.time()
                try:
                    ret = f(*args,**kw)
                    try:
                        if callable(ok):
                            self.mark(event,ok(ret))
                        elif ok is not None:
                            self.mark(event,ok)
                    except:
                        pass
                    E = time.time()
                    self.histogram(event,E-S)
                    return ret
                except Exception as exc:
                    try:
                        if callable(error):
                            self.mark(event,error(exc))
                        elif error is not None:
                            self.mark(event,error)
                    except:
                        pass   
                    E = time.time()
                    self.histogram(event,E-S)
                    raise
            if asyncio.iscoroutinefunction(f):
                return __asyncf
            else:
                return __f
        return _f

    #
    # Access to the metrics
    #
    def get_metrics(self,event=None,result=None,metric=None):
        """
        Return a structure of dictionaries with the requested metrics.

        *event*: the name of the event

        *result*: the result of the event, as used in the call to :py:meth:`mark`

        *metric*: the name of the metric, one of ``mean``, ``count``, ``rate1``, ``rate5``, ``rate15``.

        If *metric* is not provided, all metrics are returned.

        If *result* is not provided, all results are returned.

        If *event* is not provided, all events are returned.
        """
        if event:
            # if event has not yet marked anything, it is not an error
            # we may request metrics, they will be null
            if event in self._meters:
                meters_dict = self._meters[event]
            else:
                # Build a temporary empty object (do not fail - maybe the meter will exist later)
                meters_dict = collections.defaultdict(Meter)
            if result:
                if metric:
                    return getattr(meters_dict[result],metric)
                else:
                    return meters_dict[result].to_dict()
            else:
                if metric:
                    raise SyntaxError('if metric is specified, result must also be specified')
                return { k:v.to_dict() for k,v in meters_dict.items() }
        else:
            if metric or result:
                raise SyntaxError('if metric/result is specified, event must also be specified')
            return {K:{ k:v.to_dict() for k,v in V.items()} for K,V in self._meters.items()}

    def get_gauges(self,name=None,metric=None):
        """
        Return a structure of dictionaries with the gauge metrics.

        *name*: the name of the gauge

        *metric*: the name of the metric, one of ``min``, ``max``, ``count``

        If *metric* is not provided, all metrics are returned.

        If *name* is not provided, all gauges are returned.
        """
        if name:
            if name in self._gauges:
                gauge = self._gauges[name]
            else:
                # Build a temporary empty object (do not fail - maybe the gauge will exist later)
                gauge = Gauge(0)
            if metric:
                return getattr(gauge,metric)
            return gauge.to_dict()
        else:
            return { k:v.to_dict() for k,v in self._gauges.items() }

    def get_histograms(self,event=None,metric=None,percentiles=[0.05, 0.25, 0.50, 0.75, 0.95],scale=10):
        """
        Return a structure of dictionaries with the histograms metrics.

        *event*: the name of the event

        *metric*: the name of the metric, one of ``count``, ``min``, ``max``, ``mean``, ``stddev``, ``quantiles``, ``distribution``

        *percentiles*: the list of percentiles for which the quantiles are returned

        *scale*: the resolution (number of values) of the returned distribution

        If *metric* is not provided, all metrics are returned.

        If *event* is not provided, all histograms are returned.
        """
        if event:
            if event in self._histograms:
                his = self._histograms[event]
            else:
                # Build a temporary empty object (do not fail - maybe the histogram will exist later)
                his = Histogram()
            data = {}
            if not metric:
                data.update(his.to_dict())
                snapshot = his.snapshot
                data['quantiles'] = {p:snapshot.get_value(p) for p in percentiles}
                data['distribution'] = snapshot.get_distribution(scale)
                return data
            elif metric=='quantiles':
                snapshot = his.snapshot
                data = {p:snapshot.get_value(p) for p in percentiles}
            elif metric=='distribution':
                snapshot = his.snapshot
                data = snapshot.get_distribution(scale)
            else:
                data = his.to_dict()[metric]
        else:
            D = {}
            for event in self._histograms.keys():
                his = self._histograms[event]
                data = {}
                data.update(his.to_dict())
                snapshot = his.snapshot
                data['quantiles'] = {p:snapshot.get_value(p) for p in percentiles}
                data['distribution'] = snapshot.get_distribution(scale)
                D[event] = data
            data = D
        return data





import os
import unittest
import time
import threading
import http.server
import requests

from livemetrics.dashboard import dashboard

def _serve_dashboard():
    dashboard.main(['-i',os.path.join(os.path.dirname(__file__),'sample.yaml')])

from . import server
from . import client

#_______________________________________________________________________________
class TestDashboard(unittest.TestCase):
    
    def setUp(self):
        t = threading.Thread(target=_serve_dashboard, daemon=True)
        t.start()
        time.sleep(0.2)

        t = threading.Thread(target=server._serve, daemon=True)
        t.start()
        time.sleep(0.2)

        t = threading.Thread(target=client.inject, daemon=True)
        t.start()
        time.sleep(1.0)

    def test_all(self):
        # Send some events
        with requests.get('http://localhost:9000/') as r:
            self.assertEqual(200,r.status_code)

        payload = {'server': 'http://localhost:7070/monitoring/v1/is_healthy'}
        with requests.get('http://localhost:9000/all', params=payload) as r:
            self.assertEqual(200,r.status_code)

        payload = {'server': 'http://localhost:7070/monitoring/missing'}
        with requests.get('http://localhost:9000/all', params=payload) as r:
            self.assertEqual(500,r.status_code)

        payload = {'server': 'http://localhost:7070/monitoring/v1/metrics/meters/test/ok'}
        with requests.get('http://localhost:9000/all', params=payload) as r:
            self.assertEqual(200,r.status_code)
            self.assertGreater(r.json()['count'],2)

        payload = {'server': 'http://localhost:7070/monitoring/v1/metrics/histograms'}
        with requests.get('http://localhost:9000/all', params=payload) as r:
            self.assertEqual(200,r.status_code)
            self.assertGreater(r.json()['values']['count'],2)

        payload = {'server': 'http://localhost:7070/monitoring/v1/metrics/gauges'}
        with requests.get('http://localhost:9000/all', params=payload) as r:
            self.assertEqual(200,r.status_code)
            self.assertGreater(r.json()['random']['count'],0)

# ______________________________________________________________________________
if __name__=='__main__':
    unittest.main(argv=['-v'])
